/*
   This file is copyright (c) 1988 by William Couture.  All rights reserved.
*/


/*
   This version of the decode function, decodens(), is derived from the
   original.  The difference is that it is faster, as no strings are
   produced.  The reason for its existance is the Go command, which does
   not print the executing instruction, and thus does not need the strings
   produced by decode().

   Later note:  The assembler output from this function has been hand
                optimized.  If you need to change this function, please
                re-optimize the assembly code.  See the file decodens.asm
*/

/*
   Note:  The usual check for BYTE/WORD instruction is:
      if (instruction.flags & BYTE)
   The lack of a BYTE flag is taken to mean a WORD flag.
*/

#include <compiler.h>
#include <vim.h>
#include <opdefs.h>
#include <regstruc.h>

#define VALID 1  /* valid instruction decoded */
#define INVALID 0  /* invalid instruction decoded */

extern struct optable firstbyte[256];  /* from OPTABLES.C */
extern struct optable secondbyte[13][8];  /* from OPTABLES.C */
extern struct optable instruction;  /* from VIM.C */

extern struct tofrom dest;  /* from VIM.C */
extern struct tofrom source;  /* from VIM.C */

extern union regunion regs;  /* from VIM.C */

extern unsigned int prefix;  /* from VIM.C */
extern unsigned int prefixip;   /* used to hold address of executing instruction
                                 and the LAST prefix in case of interrupt
                                 during string primitive */
 
extern int aadaam_val;  /* from VIM.C, used in FNS.C */ 
extern int mod,reg,rm;  /* from VIM.C */

extern unsigned int holdofs,holdseg,holdreg;  /* defined in DECODE.C */

extern unsigned int mixbytereg[8];  /* defined in VIM.C */

extern unsigned char byte_one;  /* from VIM.C */

unsigned int fetchbyte(), fetchword();

void eabyte()  /* This routine decodes the mod,reg,rm fields of the EA byte.
                  The memory reference address is computed, and used to
                  set the dest/source VALUE field. */
   {
   long i;
   unsigned char filler,byte;
   unsigned char ch[4];

   if (mod < 3)  /* if memory addressing */
      {
      switch(rm)  /* addressing mode */
         {
         case 0:
            holdofs = regs.r.bx.word + regs.r.si;
            break;
         case 1:
            holdofs = regs.r.bx.word + regs.r.di;
            break;
         case 2:
            holdofs = regs.r.bp + regs.r.si;
            holdseg = regs.r.ss;
            break;
         case 3:
            holdofs = regs.r.bp + regs.r.di;
            holdseg = regs.r.ss;
            break;
         case 4:
            holdofs = regs.r.si;
            break;
         case 5:
            holdofs = regs.r.di;
            break;
         case 6:
            if (mod)  /* bp+disp */
               {
               holdofs = regs.r.bp;
               holdseg = regs.r.ss;
               }
            else  /* direct offset */
               {
               ch[0] = fromqueue();
               ch[1] = fromqueue();
               holdofs = (ch[1] << 8) + ch[0];
               }
            break;
         case 7:
            holdofs = regs.r.bx.word;
            break;
         }
      switch (mod)
         {
         case 1:  /* add one byte displacement, if it exists */
            byte = fromqueue();
            if (byte > 0x7f)
               holdofs += (-1 ^ 0xff) | byte;  /* that should be machine
                                              and compiler independent */
            else
               holdofs += byte;
            break;
         case 2:  /* add two byte displacement */
            ch[0] = fromqueue();
            ch[1] = fromqueue();
            i = ch[0] + (ch[1] << 8);
            holdofs += i;  /* WATCH THIS ON A MACHINE THAT DOES NOT HAVE A
                              TWO BYTE INTEGER */
            break;
         }
      }
   else  /* register addressing */
      {
      if (instruction.flags & BYTE)
         holdreg = mixbytereg[rm];
      else
         holdreg = rm;
      }
   }

void sourcedecode()
   /* NOTE: This used to be part of the decodens() procedure, but it grew
      too big for the compiler in decode(), so it was moved here as well
      to keep the two versions the same.
   */
   {
   int i;
   unsigned char filler,byte;
   unsigned char hold[6];

   switch (instruction.source)
      {
      case S8:
         source.value = (UNSIGNED_LONG)regs.bytearray[mixbytereg[reg]];
         break;
      case S16:
         source.value = (UNSIGNED_LONG)regs.wordarray[reg];
         break;
      case SDATA8:
         source.value = (UNSIGNED_LONG)fromqueue();
         break;
      case SDATA16:
         hold[0] = fromqueue();
         hold[1] = fromqueue();
         source.value = (UNSIGNED_LONG)((hold[1] << 8) + hold[0]);
         break;
      case SXDATA8:
         byte = fromqueue();
         if (byte > 0x7f)
            source.value =  (-1 ^ 0xff) | byte;  /* that should be machine
                                              and compiler independent */
         else
            source.value = (UNSIGNED_LONG)byte;
         break;
      case SOFF8:
         byte = fromqueue();
         if (byte & SIGNEDB)
            source.value = (UNSIGNED_LONG)(regs.r.ip + (0xff00 + byte));
               /* note: assumes 16-bit unsigned int */
         else
            source.value = (UNSIGNED_LONG)(regs.r.ip + byte);
         break;
      case SOFF16:
         hold[0] = fromqueue();
         hold[1] = fromqueue();
         source.value = (UNSIGNED_LONG)(regs.r.ip + ((hold[1] << 8) + hold[0]));
         break;
      case SADDR16:
         hold[0] = fromqueue();
         hold[1] = fromqueue();
         if (prefix & SEG)
            getseg();
         source.seg = holdseg;
         source.ofs = (hold[1] << 8) + hold[0];
         if (instruction.flags & BYTE)
            source.value = (UNSIGNED_LONG)fetchbyte(holdseg,source.ofs);
         else
            source.value = (UNSIGNED_LONG)fetchword(holdseg,source.ofs);
         break;
      case SADDR32:
         for (i = 0; i < 4; i++)
            hold[i] = fromqueue();
         source.ofs = (hold[1] << 8) + hold[0];
         source.seg = (hold[3] << 8) + hold[2];
           /* CALLF and JMPF don't know if their address is coming
              from immediate data or a memory reference */
         source.value = (UNSIGNED_LONG)source.seg;
         source.value = (source.value << 16) + ((UNSIGNED_LONG)source.ofs);
         break;
      case SL:
         source.where = SEGREG;
         source.reg = reg + 8;
         source.value = (UNSIGNED_LONG)regs.wordarray[source.reg];
         break;
      case SM:
         eabyte();
         if (prefix & SEG)
            getseg();
         if (instruction.flags & BYTE)
            {
            if (mod < 3)
               {
               source.seg = holdseg;
               source.ofs = holdofs;
               source.value = (UNSIGNED_LONG)fetchbyte(holdseg,holdofs);
               }
            else
               {
               source.where = REGISTER;
               source.value = (UNSIGNED_LONG)regs.bytearray[holdreg];
               source.reg = holdreg;
               }
            }
         else
            {
            if (mod < 3)
               {
               source.seg = holdseg;
               source.ofs = holdofs;
               source.value = (UNSIGNED_LONG)fetchword(holdseg,holdofs);
               if (instruction.flags & ADDR32)  /* for CALLF and JMPF */
                  {
                  source.value &= 0xffffL;
                  source.value += ((UNSIGNED_LONG)fetchword(holdseg,holdofs+2)) << 16;
                  }
               }
            else
               {
               source.where = REGISTER;
               source.value = (UNSIGNED_LONG)regs.wordarray[holdreg];
               source.reg = holdreg;
               }
            }
         break;
      }
   source.oldvalue = source.value;
   }

int decodens()
   {
   register int i;
   unsigned char filler,byte;
   unsigned char hold[6];
   
   mod = 0;  /* start off fresh */
   reg = 0;
   rm = 0;
   byte_one = fromqueue();  /* Decode first byte of instruction.
                           Structure assignment not supported in DeSmet 2.4,
                           so do individual assigns. */
/*   instruction.name = firstbyte[byte_one].name;
   instruction.flags = firstbyte[byte_one].flags;
   instruction.dest = firstbyte[byte_one].dest;
   instruction.source = firstbyte[byte_one].source; */
   instruction = firstbyte[byte_one];  /* try structure assignment */
   holdseg = regs.r.ds;  /* assume that memref's will use ds */
   if (instruction.flags & PREFIX)  /* if prefix, store info on it */
      {
      prefixip = regs.r.ip-1;  /* ip already updated in fromqueue() */
      switch (instruction.name)
         {
         case SEGCS:
            prefix |= CSSEG;
            break;
         case SEGDS:
            prefix |= DSSEG;
            break;
         case SEGES:
            prefix |= ESSEG;
            break;
         case SEGSS:
            prefix |= SSSEG;
            break;
         case REPZ:
            prefix |= REPZERO;
            break;
         case REPNZ:
            prefix |= REPNOTZERO;
            break;
         }
       }
   if ((instruction.dest < DDATA8) || (instruction.source < SDATA8))
      {  /* if EA byte addressing */
      byte = fromqueue();
      mod = (byte >> 6) & 0x03;
      reg = (byte >> 3) & 0x07;
      rm = byte & 0x07;
      if (instruction.name == UNKNOWN)
/*         if ((instruction.dest == DK) && (mod == 3) && ((reg == 3) || (reg == 5)) )
                    {* JMP FAR AX, CALL FAR DI, etc, are illegal *}
            instruction.name = UNDEF; take care of weird decode cases
         else  */
            {  /* decode second byte of instruction */
/*            instruction.name = secondbyte[instruction.dest - DA][reg].name;
            instruction.flags = secondbyte[instruction.dest - DA][reg].flags;
            instruction.source = secondbyte[instruction.dest - DA][reg].source;
            instruction.dest = secondbyte[instruction.dest - DA][reg].dest; */
            instruction = secondbyte[instruction.dest-DA][reg];
                             /* try structure assignment */
            }
      if ((reg > 3) && ((instruction.dest == DL) || (instruction.source == SL)))
         instruction.name = UNDEF;
      if ((mod == 3)
              && ((instruction.name == CALLF) || (instruction.name == JMPF)
                 || (instruction.name == BOUND) || (instruction.name == LES)
                 || (instruction.name == LDS) || (instruction.name == LEA)))
         instruction.name = UNDEF;
      }
   if (instruction.flags & AADAAM)  /* AAD and AAM are more weird cases */
      aadaam_val = fromqueue();
   if (instruction.name > DBYTE)  /* if valid instruction */
      {
      if ((instruction.dest >= DAX) && (instruction.dest <= DBH))
         {  /* handle register modes */
         if (instruction.dest >= DAL)
            {
            dest.where = REGISTER;
            dest.reg = mixbytereg[instruction.dest - DAL];
            }
         else
            if (instruction.dest >= DFL)
               {
               dest.where = REGISTER;
               if (instruction.dest == DRAL)
                  dest.reg = mixbytereg[0];
               if (instruction.dest == DRAX)
                  dest.reg  = 0;
               }
            else
               if (instruction.dest >= DES)
                  {
                  dest.where = SEGREG;
                  dest.reg = instruction.dest - DES + 8;
                  }
               else
                  {
                  dest.where = REGISTER;
                  dest.reg = instruction.dest - DAX;
                  }
         if (instruction.flags & BYTE)
            dest.value = (UNSIGNED_LONG)regs.bytearray[dest.reg];
         else
            dest.value = (UNSIGNED_LONG)regs.wordarray[dest.reg];
         dest.oldvalue = dest.value;
         }
      else
         {
         switch (instruction.dest)
            {
            case D8:
               dest.where = REGISTER;
               dest.reg = mixbytereg[reg];
               dest.value = (UNSIGNED_LONG)regs.bytearray[dest.reg];
               break;
            case D16:
               dest.where = REGISTER;
               dest.reg = reg;
               dest.value = (UNSIGNED_LONG)regs.wordarray[dest.reg];
               break;
            case DDATA8:  /* used for the OUT instruction */
               dest.value = (UNSIGNED_LONG)fromqueue();
               break;
            case DDATA16:  /* used for the ENTER instruction */
               hold[0] = fromqueue();
               hold[1] = fromqueue();
               dest.value = (UNSIGNED_LONG)((hold[1] << 8) + hold[0]);
               break;

/*          case DXDATA8:   copied for symmetry, these cases are not
                            used for destinations.  If they become
                            necessary, make sure they are current with
                            the source definitions
               break;
            case DOFF8:
               break;
            case DOFF16:
               break;
    end of comment  */

            case DADDR16:  /* 16-bit immediate addressing is used */
               hold[0] = fromqueue();
               hold[1] = fromqueue();
               if (prefix & SEG)
                  getseg();
               dest.seg = holdseg;
               dest.ofs = (hold[1] << 8) + hold[0];
               if (instruction.name != MOV)
                  if (instruction.flags & BYTE)
                     dest.value = (UNSIGNED_LONG)fetchbyte(holdseg,dest.ofs);
                  else
                     dest.value = (UNSIGNED_LONG)fetchword(holdseg,dest.ofs);
               break;

/*            case DADDR32:     back to unused again
               for (i = 0; i < 4; i++)
                  {
                  hold[i] = fromqueue();
                  }
               break;  */

            case DL:
               dest.where = SEGREG;
               dest.reg = reg + 8;
               dest.value = (UNSIGNED_LONG)regs.wordarray[dest.reg];
               break;
            case DM:
               eabyte();
               if (prefix & SEG)
                  getseg();
               if (instruction.flags & BYTE)
                  {
                  if (mod < 3)
                     {
                     dest.seg = holdseg;
                     dest.ofs = holdofs;
                     if (instruction.name != MOV)
                        dest.value = (UNSIGNED_LONG)fetchbyte(holdseg,holdofs);
                     }
                  else
                     {
                     dest.where = REGISTER;
                     dest.value = (UNSIGNED_LONG)regs.bytearray[holdreg];
                     dest.reg = holdreg;
                     }
                  }
               else
                  {
                  if (mod < 3)
                     {
                     dest.seg = holdseg;
                     dest.ofs = holdofs;
                     if (instruction.name != MOV)
                        dest.value = (UNSIGNED_LONG)fetchword(holdseg,holdofs);
                     }
                  else
                     {
                     dest.where = REGISTER;
                     dest.value = (UNSIGNED_LONG)regs.wordarray[holdreg];
                     dest.reg = holdreg;
                     }
                  }
               break;
            }
            dest.oldvalue = dest.value;
         }
      if ((instruction.source >= SAX) && (instruction.source <= SBH))
         {  /* as above, but for the source argument */
         if (instruction.source >= SAL)
            {
            source.where = REGISTER;
            source.reg = mixbytereg[instruction.dest - SAL];
            }
         else
            if (instruction.source >= SFL)
               {
               source.where = REGISTER;
               if (instruction.source == SRAL)
                  source.reg = mixbytereg[0];
               if (instruction.source == SRAX)
                  source.reg  = 0;
               }
            else
               if (instruction.source >= SES)
                  {
                  source.where = SEGREG;
                  source.reg = instruction.source - SES + 8;
                  }
               else
                  {
                  source.where = REGISTER;
                  source.reg = instruction.source - SAX;
                  }
         if (instruction.flags & BYTE)
            source.value = (UNSIGNED_LONG)regs.bytearray[source.reg];
         else
            source.value = (UNSIGNED_LONG)regs.wordarray[source.reg];
         source.oldvalue = source.value;
         }
      else
         sourcedecode();  /* see below for details */
      if ((byte_one & '\375') == '\151')  /* if IMUL immediate */
         {
                        /* put immediate data into dest */
         hold[0] = fromqueue();
         if (byte_one == '\151')  /* if 16-bit data present */
            {
            hold[1] = fromqueue();
            dest.value = (UNSIGNED_LONG)((hold[1] << 8) + hold[0]);
            }
         else
            if (hold[0] > 0x7f)
               dest.value =  (-1 ^ 0xff) | hold[0];  /* that should be machine
                                                 and compiler independent */
            else
               dest.value = (UNSIGNED_LONG)hold[0];
         }
      }
   if (instruction.name == UNDEF)
      return(INVALID);
   else
      return(VALID);
   }
